from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from typing_extensions import override

from utilities.iterables import (
    _OneStrCaseInsensitiveBijectionError,
    _OneStrCaseInsensitiveEmptyError,
    _OneStrCaseSensitiveEmptyError,
    one_str,
)

if TYPE_CHECKING:
    from collections.abc import Mapping

try:
    from enum import StrEnum as _StrEnum  # type: ignore[]
except ImportError:  # pragma: version-ge-311

    class _StrEnum(str, Enum):
        """An enum whose elements are themselves strings."""

        @staticmethod
        @override
        def _generate_next_value_(
            name: str, start: Any, count: int, last_values: Any
        ) -> str:
            _ = start, count, last_values
            return name


StrEnum = _StrEnum
_E = TypeVar("_E", bound=Enum)


def ensure_enum(
    enum: type[_E], member: _E | str, /, *, case_sensitive: bool = True
) -> _E:
    """Ensure the object is a member of the enum."""
    if isinstance(member, Enum):
        return member
    return parse_enum(enum, member, case_sensitive=case_sensitive)


def parse_enum(enum: type[_E], member: str, /, *, case_sensitive: bool = True) -> _E:
    """Parse a string into the enum."""
    names = {e.name for e in enum}
    try:
        match = one_str(names, member, case_sensitive=case_sensitive)
    except _OneStrCaseSensitiveEmptyError:
        raise _ParseEnumCaseSensitiveEmptyError(enum=enum, member=member) from None
    except _OneStrCaseInsensitiveBijectionError as error:
        raise _ParseEnumCaseInsensitiveBijectionError(
            enum=enum, member=member, counts=error.counts
        ) from None
    except _OneStrCaseInsensitiveEmptyError:
        raise _ParseEnumCaseInsensitiveEmptyError(enum=enum, member=member) from None
    return enum[match]


@dataclass(kw_only=True)
class ParseEnumError(Exception, Generic[_E]):
    enum: type[_E]
    member: str


@dataclass(kw_only=True)
class _ParseEnumCaseSensitiveEmptyError(ParseEnumError):
    @override
    def __str__(self) -> str:
        return f"Enum {self.enum} does not contain {self.member!r}."


@dataclass(kw_only=True)
class _ParseEnumCaseInsensitiveBijectionError(ParseEnumError):
    counts: Mapping[str, int]

    @override
    def __str__(self) -> str:
        return f"Enum {self.enum} must not contain duplicates (case insensitive); got {self.counts}."


@dataclass(kw_only=True)
class _ParseEnumCaseInsensitiveEmptyError(ParseEnumError):
    @override
    def __str__(self) -> str:
        return f"Enum {self.enum} does not contain {self.member!r} (case insensitive)."


__all__ = ["ParseEnumError", "StrEnum", "ensure_enum", "parse_enum"]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, unique
from platform import system
from typing import TYPE_CHECKING

from typing_extensions import assert_never, override

if TYPE_CHECKING:
    from collections.abc import Iterator

    from utilities.types import IterableStrs


@unique
class System(str, Enum):
    """An enumeration of the systems."""

    windows = "windows"
    mac = "mac"
    linux = "linux"


def get_system() -> System:
    """Get the system/OS name."""
    sys = system()
    if sys == "Windows":  # pragma: os-ne-windows
        return System.windows
    if sys == "Darwin":  # pragma: os-ne-macos
        return System.mac
    if sys == "Linux":  # pragma: os-ne-linux
        return System.linux
    raise GetSystemError(sys=sys)  # pragma: no cover


@dataclass(kw_only=True)
class GetSystemError(Exception):
    sys: str

    @override
    def __str__(self) -> str:
        return (  # pragma: no cover
            f"System must be one of Windows, Darwin, Linux; got {self.sys!r} instead"
        )


SYSTEM = get_system()
IS_WINDOWS = SYSTEM is System.windows
IS_MAC = SYSTEM is System.mac
IS_LINUX = SYSTEM is System.linux
IS_NOT_WINDOWS = not IS_WINDOWS
IS_NOT_MAC = not IS_MAC
IS_NOT_LINUX = not IS_LINUX


def maybe_yield_lower_case(text: IterableStrs, /) -> Iterator[str]:
    """Yield lower-cased text if the platform is case-insentive."""
    match SYSTEM:
        case System.windows:  # pragma: os-ne-windows
            yield from (t.lower() for t in text)
        case System.mac:  # pragma: os-ne-macos
            yield from (t.lower() for t in text)
        case System.linux:  # pragma: os-ne-linux
            yield from text
        case _ as never:  # type: ignore[]
            assert_never(never)


__all__ = [
    "IS_LINUX",
    "IS_MAC",
    "IS_NOT_LINUX",
    "IS_NOT_MAC",
    "IS_NOT_WINDOWS",
    "IS_WINDOWS",
    "SYSTEM",
    "GetSystemError",
    "System",
    "get_system",
    "maybe_yield_lower_case",
]



from __future__ import annotations

from zoneinfo import ZoneInfo

from pytest import mark, param

from utilities.zoneinfo import HONG_KONG, TOKYO, US_CENTRAL, US_EASTERN


class TestTimeZones:
    @mark.parametrize(
        "time_zone",
        [param(HONG_KONG), param(TOKYO), param(US_CENTRAL), param(US_EASTERN)],
    )
    def test_main(self, *, time_zone: ZoneInfo) -> None:
        assert isinstance(time_zone, ZoneInfo)



from __future__ import annotations

from operator import itemgetter
from time import sleep
from typing import TYPE_CHECKING

from pytest import raises

from tests.conftest import FLAKY
from utilities.cacher import cache_to_disk
from utilities.functions import identity
from utilities.iterables import one
from utilities.types import EnsureClassError

if TYPE_CHECKING:
    from pathlib import Path


class TestCacheToDisk:
    def test_main(self, *, tmp_path: Path) -> None:
        counter = 0

        @cache_to_disk(root=tmp_path)
        def func(x: int, /) -> int:
            nonlocal counter
            counter += 1
            return x

        assert len(list(tmp_path.iterdir())) == 0
        assert func(0) == 0
        path = one(one(tmp_path.iterdir()).iterdir())
        assert path.name == "func"
        assert len(list(path.iterdir())) == counter == 1
        assert func(0) == 0
        assert len(list(path.iterdir())) == counter == 1
        for _ in range(2):
            assert func(1) == 1
            assert len(list(path.iterdir())) == counter == 2
        for _ in range(2):
            assert func(2) == 2
            assert len(list(path.iterdir())) == 3

    def test_max_size(self, *, tmp_path: Path) -> None:
        func = cache_to_disk(root=tmp_path, max_size=2)(identity)
        assert func(0) == 0
        path = one(one(tmp_path.iterdir()).iterdir())
        assert len(list(path.iterdir())) == 1
        sleep(duration := 0.01)
        assert func(1) == 1
        assert len(list(path.iterdir())) == 2
        mapping = {p: p.stat().st_mtime for p in path.iterdir()}
        as_list = sorted(mapping.items(), key=itemgetter(1))
        pre_old, pre_new = (time for _, time in as_list)
        sleep(duration)
        assert func(2) == 2
        assert len(list(path.iterdir())) == 2
        mapping = {p: p.stat().st_mtime for p in path.iterdir()}
        as_list = sorted(mapping.items(), key=itemgetter(1))
        post_old, post_new = (time for _, time in as_list)
        assert pre_old < pre_new == post_old < post_new

    def test_skip(self, *, tmp_path: Path) -> None:
        func = cache_to_disk(root=tmp_path, skip=True)(identity)
        assert len(list(tmp_path.iterdir())) == 0
        assert func(0) == 0
        assert len(list(tmp_path.iterdir())) == 0

    @FLAKY
    def test_max_duration(self, *, tmp_path: Path) -> None:
        max_duration = 0.1
        func = cache_to_disk(root=tmp_path, max_duration=max_duration)(identity)
        assert func(0) == 0
        path = one(one(tmp_path.iterdir()).iterdir())
        orig = path.stat().st_mtime
        assert func(0) == 0
        assert path.stat().st_mtime == orig
        sleep(2 * max_duration)
        assert func(1) == 1
        assert path.stat().st_mtime > orig

    def test_args_and_kwargs_resolved(self, *, tmp_path: Path) -> None:
        @cache_to_disk(root=tmp_path)
        def add(x: int, y: int) -> int:
            return x + y

        assert add(0, 0) == 0
        path = one(one(tmp_path.iterdir()).iterdir())
        assert len(list(path.iterdir())) == 1
        assert add(0, y=0) == 0
        assert len(list(path.iterdir())) == 1
        assert add(x=0, y=0) == 0
        assert len(list(path.iterdir())) == 1
        assert add(x=1, y=1) == 2
        assert len(list(path.iterdir())) == 2

    def test_rerun(self, *, tmp_path: Path) -> None:
        func = cache_to_disk(root=tmp_path)(identity)
        assert func(0) == 0
        path = one(one(tmp_path.iterdir()).iterdir())
        assert len(list(path.iterdir())) == 1
        pre = path.stat().st_mtime
        sleep(0.01)
        assert func(0, rerun=True) == 0  # type: ignore[]
        post = path.stat().st_mtime
        assert pre < post

    def test_rerun_not_a_boolean(self, *, tmp_path: Path) -> None:
        func = cache_to_disk(root=tmp_path)(identity)
        with raises(EnsureClassError):
            _ = func(0, rerun=None)  # type: ignore[]



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Sequence
from dataclasses import dataclass
from itertools import islice
from textwrap import indent
from typing import Any, Generic, Literal, TypeVar, cast, overload

from more_itertools import always_iterable as _always_iterable
from more_itertools import peekable as _peekable
from more_itertools import split_into
from more_itertools import windowed_complete as _windowed_complete
from typing_extensions import assert_never, override

from utilities.sentinel import Sentinel, sentinel

_T = TypeVar("_T")
_U = TypeVar("_U")


def always_iterable(
    obj: _T | Iterable[_T],
    /,
    *,
    base_type: type[Any] | tuple[type[Any], ...] | None = (str, bytes),
) -> Iterator[_T]:
    """Typed version of `always_iterable`."""
    return _always_iterable(obj, base_type=base_type)


@overload
def filter_include_and_exclude(
    iterable: Iterable[_T],
    /,
    *,
    include: _U | Iterable[_U] | None = None,
    exclude: _U | Iterable[_U] | None = None,
    key: Callable[[_T], _U],
) -> Iterable[_T]: ...
@overload
def filter_include_and_exclude(
    iterable: Iterable[_T],
    /,
    *,
    include: _T | Iterable[_T] | None = None,
    exclude: _T | Iterable[_T] | None = None,
    key: Callable[[_T], Any] | None = None,
) -> Iterable[_T]: ...
def filter_include_and_exclude(
    iterable: Iterable[_T],
    /,
    *,
    include: _U | Iterable[_U] | None = None,
    exclude: _U | Iterable[_U] | None = None,
    key: Callable[[_T], _U] | None = None,
) -> Iterable[_T]:
    """Filter an iterable based on an inclusion/exclusion pair."""
    include, exclude = resolve_include_and_exclude(include=include, exclude=exclude)
    if include is not None:
        if key is None:
            iterable = (x for x in iterable if x in include)
        else:
            iterable = (x for x in iterable if key(x) in include)
    if exclude is not None:
        if key is None:
            iterable = (x for x in iterable if x not in exclude)
        else:
            iterable = (x for x in iterable if key(x) not in exclude)
    return iterable


class peekable(_peekable, Generic[_T]):  # noqa: N801
    """Peekable which supports dropwhile/takewhile methods."""

    def __init__(self, iterable: Iterable[_T], /) -> None:
        super().__init__(iterable)

    @override
    def __next__(self) -> _T:
        return super().__next__()

    def dropwhile(self, predicate: Callable[[_T], bool], /) -> None:
        while bool(self) and predicate(self.peek()):
            _ = next(self)

    @overload
    def peek(self, *, default: Sentinel = sentinel) -> _T: ...
    @overload
    def peek(self, *, default: _U) -> _T | _U: ...
    @override
    def peek(self, *, default: Any = sentinel) -> Any:  # type: ignore[]
        if isinstance(default, Sentinel):
            return super().peek()
        return super().peek(default=default)

    def takewhile(self, predicate: Callable[[_T], bool], /) -> Iterator[_T]:
        while bool(self) and predicate(self.peek()):
            yield next(self)


def resolve_include_and_exclude(
    *,
    include: _T | Iterable[_T] | None = None,
    exclude: _T | Iterable[_T] | None = None,
) -> tuple[set[_T] | None, set[_T] | None]:
    """Resolve an inclusion/exclusion pair."""
    include_use = include if include is None else set(always_iterable(include))
    exclude_use = exclude if exclude is None else set(always_iterable(exclude))
    if (
        (include_use is not None)
        and (exclude_use is not None)
        and (len(include_use & exclude_use) >= 1)
    ):
        raise ResolveIncludeAndExcludeError(include=include_use, exclude=exclude_use)
    return include_use, exclude_use


@dataclass(kw_only=True)
class ResolveIncludeAndExcludeError(Exception, Generic[_T]):
    include: Iterable[_T]
    exclude: Iterable[_T]

    @override
    def __str__(self) -> str:
        include = list(self.include)
        exclude = list(self.exclude)
        overlap = set(include) & set(exclude)
        return f"Iterables {include} and {exclude} must not overlap; got {overlap}."


def windowed_complete(
    iterable: Iterable[_T], n: int, /
) -> Iterator[tuple[tuple[_T, ...], tuple[_T, ...], tuple[_T, ...]]]:
    """Typed version of `windowed_complete`."""
    return cast(
        Iterator[tuple[tuple[_T, ...], tuple[_T, ...], tuple[_T, ...]]],
        _windowed_complete(iterable, n),
    )


@dataclass(frozen=True)
class Split(Generic[_T]):
    """An iterable split into head/tail."""

    head: _T
    tail: _T

    @override
    def __repr__(self) -> str:
        spaces = 4 * " "
        head_first = indent("head=", spaces)
        head_rest = indent(repr(self.head), 2 * spaces)
        tail_first = indent("tail=", spaces)
        tail_rest = indent(repr(self.tail), 2 * spaces)
        joined = f"{head_first}\n{head_rest}\n{tail_first}\n{tail_rest}"
        return f"{type(self).__name__}(\n{joined}\n)"


def yield_splits(
    iterable: Iterable[_T],
    head: int,
    tail: int,
    /,
    *,
    min_frac: float | None = None,
    freq: int | None = None,
) -> Iterator[Split[Sequence[_T]]]:
    """Yield the splits of an iterable."""
    it1 = _yield_splits1(iterable, head + tail)
    it2 = _yield_splits2(it1, head, tail, min_frac=min_frac)
    it3 = _yield_splits3(it2)
    freq_use = tail if freq is None else freq
    return islice(it3, 0, None, freq_use)


def _yield_splits1(
    iterable: Iterable[_T], total: int, /
) -> Iterator[tuple[Literal["head", "body"], Sequence[_T]]]:
    peek = peekable(iterable)
    for i in range(1, total + 1):
        if len(result := peek[:i]) < i:
            return
        yield "head", result
    while True:
        _ = next(peek)
        if len(result := peek[:total]) >= 1:
            yield "body", result
        else:
            break


def _yield_splits2(
    iterable: Iterable[tuple[Literal["head", "body"], Sequence[_T]],],
    head: int,
    tail: int,
    /,
    *,
    min_frac: float | None = None,
) -> Iterator[tuple[Iterable[_T], int, int]]:
    min_length = head if min_frac is None else min_frac * head
    for kind, window in iterable:
        len_win = len(window)
        match kind:
            case "head":
                len_head = max(len_win - tail, 0)
                if len_head >= min_length:
                    yield window, len_head, tail
            case "body":
                len_tail = max(len_win - head, 0)
                if len_tail >= 1:
                    yield window, head, len_tail
            case _ as never:  # type: ignore[]
                assert_never(never)


def _yield_splits3(
    iterable: Iterable[tuple[Iterable[_T], int, int]], /
) -> Iterator[Split[Sequence[_T]]]:
    for window, len_head, len_tail in iterable:
        head_win, tail_win = split_into(window, [len_head, len_tail])
        yield Split(head=list(head_win), tail=list(tail_win))


__all__ = [
    "ResolveIncludeAndExcludeError",
    "Split",
    "always_iterable",
    "filter_include_and_exclude",
    "peekable",
    "resolve_include_and_exclude",
    "windowed_complete",
    "yield_splits",
]



from __future__ import annotations

import datetime as dt
from contextlib import contextmanager
from dataclasses import dataclass
from functools import reduce
from itertools import repeat
from typing import TYPE_CHECKING, Annotated, Any, Literal, cast, overload

import numpy as np
from numpy import (
    array,
    bool_,
    datetime64,
    digitize,
    dtype,
    errstate,
    exp,
    flatnonzero,
    flip,
    float64,
    full_like,
    inf,
    int64,
    isclose,
    isdtype,
    isfinite,
    isinf,
    isnan,
    linspace,
    log,
    nan,
    nanquantile,
    ndarray,
    object_,
    prod,
    rint,
    roll,
    where,
)
from numpy.linalg import det, eig
from numpy.random import default_rng
from numpy.typing import NDArray
from typing_extensions import override

from utilities.datetime import EPOCH_UTC, UTC
from utilities.errors import redirect_error
from utilities.iterables import is_iterable_not_str

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from utilities.math import FloatFinPos

# RNG
DEFAULT_RNG = default_rng()


# types
Datetime64Unit = Literal[
    "Y", "M", "W", "D", "h", "m", "s", "ms", "us", "ns", "ps", "fs", "as"
]
Datetime64Kind = Literal["date", "time"]


# dtypes
datetime64Y = dtype("datetime64[Y]")  # noqa: N816
datetime64M = dtype("datetime64[M]")  # noqa: N816
datetime64W = dtype("datetime64[W]")  # noqa: N816
datetime64D = dtype("datetime64[D]")  # noqa: N816
datetime64h = dtype("datetime64[h]")
datetime64m = dtype("datetime64[m]")
datetime64s = dtype("datetime64[s]")
datetime64ms = dtype("datetime64[ms]")
datetime64us = dtype("datetime64[us]")
datetime64ns = dtype("datetime64[ns]")
datetime64ps = dtype("datetime64[ps]")
datetime64fs = dtype("datetime64[fs]")
datetime64as = dtype("datetime64[as]")


timedelta64Y = dtype("timedelta64[Y]")  # noqa: N816
timedelta64M = dtype("timedelta64[M]")  # noqa: N816
timedelta64W = dtype("timedelta64[W]")  # noqa: N816
timedelta64D = dtype("timedelta64[D]")  # noqa: N816
timedelta64h = dtype("timedelta64[h]")
timedelta64m = dtype("timedelta64[m]")
timedelta64s = dtype("timedelta64[s]")
timedelta64ms = dtype("timedelta64[ms]")
timedelta64us = dtype("timedelta64[us]")
timedelta64ns = dtype("timedelta64[ns]")
timedelta64ps = dtype("timedelta64[ps]")
timedelta64fs = dtype("timedelta64[fs]")
timedelta64as = dtype("timedelta64[as]")


# annotations - dtypes
NDArrayA = NDArray[Any]
NDArrayB = NDArray[bool_]
NDArrayF = NDArray[float64]
NDArrayI = NDArray[int64]
NDArrayO = NDArray[object_]
NDArrayDY = Annotated[NDArrayA, datetime64Y]
NDArrayDM = Annotated[NDArrayA, datetime64M]
NDArrayDW = Annotated[NDArrayA, datetime64W]
NDArrayDD = Annotated[NDArrayA, datetime64D]
NDArrayDh = Annotated[NDArrayA, datetime64h]
NDArrayDm = Annotated[NDArrayA, datetime64m]
NDArrayDs = Annotated[NDArrayA, datetime64s]
NDArrayDms = Annotated[NDArrayA, datetime64ms]
NDArrayDus = Annotated[NDArrayA, datetime64us]
NDArrayDns = Annotated[NDArrayA, datetime64ns]
NDArrayDps = Annotated[NDArrayA, datetime64ps]
NDArrayDfs = Annotated[NDArrayA, datetime64fs]
NDArrayDas = Annotated[NDArrayA, datetime64as]
NDArrayD = (
    NDArrayDY
    | NDArrayDM
    | NDArrayDW
    | NDArrayDD
    | NDArrayDh
    | NDArrayDm
    | NDArrayDs
    | NDArrayDms
    | NDArrayDus
    | NDArrayDns
    | NDArrayDps
    | NDArrayDfs
    | NDArrayDas
)


# annotations - ndims
NDArray0 = Annotated[NDArrayA, 0]
NDArray1 = Annotated[NDArrayA, 1]
NDArray2 = Annotated[NDArrayA, 2]
NDArray3 = Annotated[NDArrayA, 3]


# annotations - dtype & ndim
NDArrayB0 = Annotated[NDArrayB, 0]
NDArrayD0 = Annotated[NDArrayD, 0]
NDArrayF0 = Annotated[NDArrayF, 0]
NDArrayI0 = Annotated[NDArrayI, 0]
NDArrayO0 = Annotated[NDArrayO, 0]
NDArrayDY0 = Annotated[NDArrayDY, 0]
NDArrayDM0 = Annotated[NDArrayDM, 0]
NDArrayDW0 = Annotated[NDArrayDW, 0]
NDArrayDD0 = Annotated[NDArrayDD, 0]
NDArrayDh0 = Annotated[NDArrayDh, 0]
NDArrayDm0 = Annotated[NDArrayDm, 0]
NDArrayDs0 = Annotated[NDArrayDs, 0]
NDArrayDms0 = Annotated[NDArrayDms, 0]
NDArrayDus0 = Annotated[NDArrayDus, 0]
NDArrayDns0 = Annotated[NDArrayDns, 0]
NDArrayDps0 = Annotated[NDArrayDps, 0]
NDArrayDfs0 = Annotated[NDArrayDfs, 0]
NDArrayDas0 = Annotated[NDArrayDas, 0]

NDArrayB1 = Annotated[NDArrayB, 1]
NDArrayD1 = Annotated[NDArrayD, 1]
NDArrayF1 = Annotated[NDArrayF, 1]
NDArrayI1 = Annotated[NDArrayI, 1]
NDArrayO1 = Annotated[NDArrayO, 1]
NDArrayDY1 = Annotated[NDArrayDY, 1]
NDArrayDM1 = Annotated[NDArrayDM, 1]
NDArrayDW1 = Annotated[NDArrayDW, 1]
NDArrayDD1 = Annotated[NDArrayDD, 1]
NDArrayDh1 = Annotated[NDArrayDh, 1]
NDArrayDm1 = Annotated[NDArrayDm, 1]
NDArrayDs1 = Annotated[NDArrayDs, 1]
NDArrayDms1 = Annotated[NDArrayDms, 1]
NDArrayDus1 = Annotated[NDArrayDus, 1]
NDArrayDns1 = Annotated[NDArrayDns, 1]
NDArrayDps1 = Annotated[NDArrayDps, 1]
NDArrayDfs1 = Annotated[NDArrayDfs, 1]
NDArrayDas1 = Annotated[NDArrayDas, 1]

NDArrayB2 = Annotated[NDArrayB, 2]
NDArrayD2 = Annotated[NDArrayD, 2]
NDArrayF2 = Annotated[NDArrayF, 2]
NDArrayI2 = Annotated[NDArrayI, 2]
NDArrayO2 = Annotated[NDArrayO, 2]
NDArrayDY2 = Annotated[NDArrayDY, 2]
NDArrayDM2 = Annotated[NDArrayDM, 2]
NDArrayDW2 = Annotated[NDArrayDW, 2]
NDArrayDD2 = Annotated[NDArrayDD, 2]
NDArrayDh2 = Annotated[NDArrayDh, 2]
NDArrayDm2 = Annotated[NDArrayDm, 2]
NDArrayDs2 = Annotated[NDArrayDs, 2]
NDArrayDms2 = Annotated[NDArrayDms, 2]
NDArrayDus2 = Annotated[NDArrayDus, 2]
NDArrayDns2 = Annotated[NDArrayDns, 2]
NDArrayDps2 = Annotated[NDArrayDps, 2]
NDArrayDfs2 = Annotated[NDArrayDfs, 2]
NDArrayDas2 = Annotated[NDArrayDas, 2]

NDArrayB3 = Annotated[NDArrayB, 3]
NDArrayD3 = Annotated[NDArrayD, 3]
NDArrayF3 = Annotated[NDArrayF, 3]
NDArrayI3 = Annotated[NDArrayI, 3]
NDArrayO3 = Annotated[NDArrayO, 3]
NDArrayDY3 = Annotated[NDArrayDY, 3]
NDArrayDM3 = Annotated[NDArrayDM, 3]
NDArrayDW3 = Annotated[NDArrayDW, 3]
NDArrayDD3 = Annotated[NDArrayDD, 3]
NDArrayDh3 = Annotated[NDArrayDh, 3]
NDArrayDm3 = Annotated[NDArrayDm, 3]
NDArrayDs3 = Annotated[NDArrayDs, 3]
NDArrayDms3 = Annotated[NDArrayDms, 3]
NDArrayDus3 = Annotated[NDArrayDus, 3]
NDArrayDns3 = Annotated[NDArrayDns, 3]
NDArrayDps3 = Annotated[NDArrayDps, 3]
NDArrayDfs3 = Annotated[NDArrayDfs, 3]
NDArrayDas3 = Annotated[NDArrayDas, 3]


# functions


def array_indexer(i: int, ndim: int, /, *, axis: int = -1) -> tuple[int | slice, ...]:
    """Get the indexer which returns the `ith` slice of an array along an axis."""
    indexer: list[int | slice] = list(repeat(slice(None), times=ndim))
    indexer[axis] = i
    return tuple(indexer)


def as_int(
    array: NDArrayF, /, *, nan: int | None = None, inf: int | None = None
) -> NDArrayI:
    """Safely cast an array of floats into ints."""
    if (is_nan := isnan(array)).any():
        if nan is None:
            msg = f"{array=}"
            raise AsIntError(msg)
        return as_int(where(is_nan, nan, array).astype(float))
    if (is_inf := isinf(array)).any():
        if inf is None:
            msg = f"{array=}"
            raise AsIntError(msg)
        return as_int(where(is_inf, inf, array).astype(float))
    rounded = rint(array)
    if (isfinite(array) & (~isclose(array, rounded))).any():
        msg = f"{array=}"
        raise AsIntError(msg)
    return rounded.astype(int)


class AsIntError(Exception): ...


def date_to_datetime64(date: dt.date, /) -> datetime64:
    """Convert a `dt.date` to `numpy.datetime64`."""
    return datetime64(date, "D")


DATE_MIN_AS_DATETIME64 = date_to_datetime64(dt.date.min)
DATE_MAX_AS_DATETIME64 = date_to_datetime64(dt.date.max)


def datetime_to_datetime64(datetime: dt.datetime, /) -> datetime64:
    """Convert a `dt.datetime` to `numpy.datetime64`."""
    if (tz := datetime.tzinfo) is None:
        datetime_use = datetime
    elif tz is UTC:
        datetime_use = datetime.replace(tzinfo=None)
    else:
        raise DatetimeToDatetime64Error(datetime=datetime, tzinfo=tz)
    return datetime64(datetime_use, "us")


@dataclass(kw_only=True)
class DatetimeToDatetime64Error(Exception):
    datetime: dt.datetime
    tzinfo: dt.tzinfo

    @override
    def __str__(self) -> str:
        return (  # pragma: no cover
            f"Timezone must be None or UTC; got {self.tzinfo}."
        )


DATETIME_MIN_AS_DATETIME64 = datetime_to_datetime64(dt.datetime.min)
DATETIME_MAX_AS_DATETIME64 = datetime_to_datetime64(dt.datetime.max)


def datetime64_to_date(datetime: datetime64, /) -> dt.date:
    """Convert a `numpy.datetime64` to a `dt.date`."""
    as_int = datetime64_to_int(datetime)
    if (dtype := datetime.dtype) == datetime64D:
        with redirect_error(
            OverflowError, DateTime64ToDateError(f"{datetime=}, {dtype=}")
        ):
            return (EPOCH_UTC + dt.timedelta(days=as_int)).date()
    msg = f"{datetime=}, {dtype=}"
    raise NotImplementedError(msg)


class DateTime64ToDateError(Exception): ...


def datetime64_to_int(datetime: datetime64, /) -> int:
    """Convert a `numpy.datetime64` to an `int`."""
    return datetime.astype(int64).item()


DATE_MIN_AS_INT = datetime64_to_int(DATE_MIN_AS_DATETIME64)
DATE_MAX_AS_INT = datetime64_to_int(DATE_MAX_AS_DATETIME64)
DATETIME_MIN_AS_INT = datetime64_to_int(DATETIME_MIN_AS_DATETIME64)
DATETIME_MAX_AS_INT = datetime64_to_int(DATETIME_MAX_AS_DATETIME64)


def datetime64_to_datetime(datetime: datetime64, /) -> dt.datetime:
    """Convert a `numpy.datetime64` to a `dt.datetime`."""
    as_int = datetime64_to_int(datetime)
    if (dtype := datetime.dtype) == datetime64ms:
        with redirect_error(
            OverflowError, DateTime64ToDateTimeError(f"{datetime=}, {dtype=}")
        ):
            return EPOCH_UTC + dt.timedelta(milliseconds=as_int)
    if dtype == datetime64us:
        return EPOCH_UTC + dt.timedelta(microseconds=as_int)
    if dtype == datetime64ns:
        microseconds, nanoseconds = divmod(as_int, int(1e3))
        if nanoseconds != 0:
            msg = f"{datetime=}, {nanoseconds=}"
            raise DateTime64ToDateTimeError(msg)
        return EPOCH_UTC + dt.timedelta(microseconds=microseconds)
    msg = f"{datetime=}, {dtype=}"
    raise NotImplementedError(msg)


class DateTime64ToDateTimeError(Exception): ...


def discretize(x: NDArrayF1, bins: int | Iterable[float], /) -> NDArrayF1:
    """Discretize an array of floats.

    Finite values are mapped to {0, ..., bins-1}.
    """
    if len(x) == 0:
        return array([], dtype=float)
    if isinstance(bins, int):
        bins_use = linspace(0, 1, num=bins + 1)
    else:
        bins_use = array(list(bins), dtype=float)
    if (is_fin := isfinite(x)).all():
        edges = nanquantile(x, bins_use)
        edges[[0, -1]] = [-inf, inf]
        return digitize(x, edges[1:]).astype(float)
    out = full_like(x, nan, dtype=float)
    out[is_fin] = discretize(x[is_fin], bins)
    return out


def ewma(array: NDArrayF, halflife: FloatFinPos, /, *, axis: int = -1) -> NDArrayF:
    """Compute the EWMA of an array."""
    from numbagg import move_exp_nanmean

    alpha = _exp_weighted_alpha(halflife)
    return cast(Any, move_exp_nanmean)(array, axis=axis, alpha=alpha)


def exp_moving_sum(
    array: NDArrayF, halflife: FloatFinPos, /, *, axis: int = -1
) -> NDArrayF:
    """Compute the exponentially-weighted moving sum of an array."""
    from numbagg import move_exp_nansum

    alpha = _exp_weighted_alpha(halflife)
    return cast(Any, move_exp_nansum)(array, axis=axis, alpha=alpha)


def _exp_weighted_alpha(halflife: FloatFinPos, /) -> float:
    """Get the alpha."""
    decay = 1.0 - exp(log(0.5) / halflife)
    com = 1.0 / decay - 1.0
    return 1.0 / (1.0 + com)


def ffill(array: NDArrayF, /, *, limit: int | None = None, axis: int = -1) -> NDArrayF:
    """Forward fill the elements in an array."""
    from bottleneck import push

    return push(array, n=limit, axis=axis)


def ffill_non_nan_slices(
    array: NDArrayF, /, *, limit: int | None = None, axis: int = -1
) -> NDArrayF:
    """Forward fill the slices in an array which contain non-nan values."""
    ndim = array.ndim
    arrays = (
        array[array_indexer(i, ndim, axis=axis)] for i in range(array.shape[axis])
    )
    out = array.copy()
    for i, repl_i in _ffill_non_nan_slices_helper(arrays, limit=limit):
        out[array_indexer(i, ndim, axis=axis)] = repl_i
    return out


def _ffill_non_nan_slices_helper(
    arrays: Iterator[NDArrayF], /, *, limit: int | None = None
) -> Iterator[tuple[int, NDArrayF]]:
    """Yield the slices to be pasted in."""
    last: tuple[int, NDArrayF] | None = None
    for i, arr_i in enumerate(arrays):
        if (~isnan(arr_i)).any():
            last = i, arr_i
        elif last is not None:
            last_i, last_sl = last
            if (limit is None) or ((i - last_i) <= limit):
                yield i, last_sl


def fillna(array: NDArrayF, /, *, value: float = 0.0) -> NDArrayF:
    """Fill the null elements in an array."""
    return where(isnan(array), value, array)


def flatn0(array: NDArrayB, /) -> int:
    """Return the index of the unique True element."""
    if not array.any():
        raise FlatN0EmptyError(array=array)
    flattened = flatnonzero(array)
    try:
        return flattened.item()
    except ValueError:
        raise FlatN0MultipleError(array=array) from None


@dataclass(kw_only=True)
class FlatN0Error(Exception):
    array: NDArrayB


@dataclass(kw_only=True)
class FlatN0EmptyError(FlatN0Error):
    @override
    def __str__(self) -> str:
        return f"Array {self.array} must contain a True."


@dataclass(kw_only=True)
class FlatN0MultipleError(FlatN0Error):
    @override
    def __str__(self) -> str:
        return f"Array {self.array} must contain at most one True."


def get_fill_value(dtype_: Any, /) -> Any:
    """Get the default fill value for a given dtype."""
    try:
        dtype_use = dtype(dtype_)
    except TypeError:
        raise GetFillValueError(dtype_=dtype_) from None
    if isdtype(dtype_use, bool_):
        return False
    if isdtype(dtype_use, (datetime64D, datetime64Y, datetime64ns)):
        return datetime64("NaT")
    if isdtype(dtype_use, float64):
        return nan
    if isdtype(dtype_use, int64):
        return 0
    return None


@dataclass(kw_only=True)
class GetFillValueError(Exception):
    dtype_: Any

    @override
    def __str__(self) -> str:
        return f"Invalid data type; got {self.dtype_!r}"


def has_dtype(x: Any, dtype: Any, /) -> bool:
    """Check if an object has the required dtype."""
    if is_iterable_not_str(dtype):
        return any(has_dtype(x, d) for d in dtype)
    return x.dtype == dtype


def is_at_least(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x >= y."""
    return (x >= y) | _is_close(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan)


def is_at_least_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x >= y or x == nan."""
    return is_at_least(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


def is_at_most(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x <= y."""
    return (x <= y) | _is_close(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan)


def is_at_most_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x <= y or x == nan."""
    return is_at_most(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


def is_between(
    x: Any,
    low: Any,
    high: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
    low_equal_nan: bool = False,
    high_equal_nan: bool = False,
) -> Any:
    """Check if low <= x <= high."""
    return is_at_least(
        x, low, rtol=rtol, atol=atol, equal_nan=equal_nan or low_equal_nan
    ) & is_at_most(x, high, rtol=rtol, atol=atol, equal_nan=equal_nan or high_equal_nan)


def is_between_or_nan(
    x: Any,
    low: Any,
    high: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
    low_equal_nan: bool = False,
    high_equal_nan: bool = False,
) -> Any:
    """Check if low <= x <= high or x == nan."""
    return is_between(
        x,
        low,
        high,
        rtol=rtol,
        atol=atol,
        equal_nan=equal_nan,
        low_equal_nan=low_equal_nan,
        high_equal_nan=high_equal_nan,
    ) | isnan(x)


def _is_close(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x == y."""
    return np.isclose(
        x,
        y,
        **({} if rtol is None else {"rtol": rtol}),
        **({} if atol is None else {"atol": atol}),
        equal_nan=equal_nan,
    )


def is_empty(shape_or_array: int | tuple[int, ...] | NDArrayA, /) -> bool:
    """Check if an ndarray is empty."""
    if isinstance(shape_or_array, int):
        return shape_or_array == 0
    if isinstance(shape_or_array, tuple):
        return (len(shape_or_array) == 0) or (prod(shape_or_array).item() == 0)
    return is_empty(shape_or_array.shape)


def is_finite_and_integral(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < inf and x == int(x)."""
    return isfinite(x) & is_integral(x, rtol=rtol, atol=atol)


def is_finite_and_integral_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < inf and x == int(x), or x == nan."""
    return is_finite_and_integral(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_negative(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < 0."""
    return isfinite(x) & is_negative(x, rtol=rtol, atol=atol)


def is_finite_and_negative_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < 0 or x == nan."""
    return is_finite_and_negative(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_non_negative(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if 0 <= x < inf."""
    return isfinite(x) & is_non_negative(x, rtol=rtol, atol=atol)


def is_finite_and_non_negative_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if 0 <= x < inf or x == nan."""
    return is_finite_and_non_negative(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_non_positive(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x <= 0."""
    return isfinite(x) & is_non_positive(x, rtol=rtol, atol=atol)


def is_finite_and_non_positive_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x <= 0 or x == nan."""
    return is_finite_and_non_positive(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_non_zero(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if -inf < x < inf, x != 0."""
    return isfinite(x) & is_non_zero(x, rtol=rtol, atol=atol)


def is_finite_and_non_zero_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x != 0 or x == nan."""
    return is_finite_and_non_zero(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_and_positive(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if 0 < x < inf."""
    return isfinite(x) & is_positive(x, rtol=rtol, atol=atol)


def is_finite_and_positive_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if 0 < x < inf or x == nan."""
    return is_finite_and_positive(x, rtol=rtol, atol=atol) | isnan(x)


def is_finite_or_nan(x: Any, /) -> Any:
    """Check if -inf < x < inf or x == nan."""
    return isfinite(x) | isnan(x)


def is_greater_than(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x > y."""
    return ((x > y) & ~_is_close(x, y, rtol=rtol, atol=atol)) | (
        equal_nan & isnan(x) & isnan(y)
    )


def is_greater_than_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x > y or x == nan."""
    return is_greater_than(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


def is_integral(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == int(x)."""
    return _is_close(x, rint(x), rtol=rtol, atol=atol)


def is_integral_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == int(x) or x == nan."""
    return is_integral(x, rtol=rtol, atol=atol) | isnan(x)


def is_less_than(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x < y."""
    return ((x < y) & ~_is_close(x, y, rtol=rtol, atol=atol)) | (
        equal_nan & isnan(x) & isnan(y)
    )


def is_less_than_or_nan(
    x: Any,
    y: Any,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> Any:
    """Check if x < y or x == nan."""
    return is_less_than(x, y, rtol=rtol, atol=atol, equal_nan=equal_nan) | isnan(x)


def is_negative(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x < 0."""
    return is_less_than(x, 0.0, rtol=rtol, atol=atol)


def is_negative_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x < 0 or x == nan."""
    return is_negative(x, rtol=rtol, atol=atol) | isnan(x)


def is_non_empty(shape_or_array: int | tuple[int, ...] | NDArrayA, /) -> bool:
    """Check if an ndarray is non-empty."""
    if isinstance(shape_or_array, int):
        return shape_or_array >= 1
    if isinstance(shape_or_array, tuple):
        return (len(shape_or_array) >= 1) and (prod(shape_or_array).item() >= 1)
    return is_non_empty(shape_or_array.shape)


def is_non_negative(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x >= 0."""
    return is_at_least(x, 0.0, rtol=rtol, atol=atol)


def is_non_negative_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x >= 0 or x == nan."""
    return is_non_negative(x, rtol=rtol, atol=atol) | isnan(x)


def is_non_positive(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x <= 0."""
    return is_at_most(x, 0.0, rtol=rtol, atol=atol)


def is_non_positive_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x <=0 or x == nan."""
    return is_non_positive(x, rtol=rtol, atol=atol) | isnan(x)


def is_non_singular(
    array: NDArrayF2 | NDArrayI2,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
) -> bool:
    """Check if det(x) != 0."""
    try:
        with errstate(over="raise"):
            return is_non_zero(det(array), rtol=rtol, atol=atol).item()
    except FloatingPointError:  # pragma: no cover
        return False


def is_non_zero(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x != 0."""
    return ~_is_close(x, 0.0, rtol=rtol, atol=atol)


def is_non_zero_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x != 0 or x == nan."""
    return is_non_zero(x, rtol=rtol, atol=atol) | isnan(x)


def is_positive(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x > 0."""
    return is_greater_than(x, 0, rtol=rtol, atol=atol)


def is_positive_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x > 0 or x == nan."""
    return is_positive(x, rtol=rtol, atol=atol) | isnan(x)


def is_positive_semidefinite(x: NDArrayF2 | NDArrayI2, /) -> bool:
    """Check if `x` is positive semidefinite."""
    if not is_symmetric(x):
        return False
    w, _ = eig(x)
    return bool(is_non_negative(w).all())


def is_symmetric(
    array: NDArrayF2 | NDArrayI2,
    /,
    *,
    rtol: float | None = None,
    atol: float | None = None,
    equal_nan: bool = False,
) -> bool:
    """Check if x == x.T."""
    m, n = array.shape
    return (m == n) and (
        _is_close(array, array.T, rtol=rtol, atol=atol, equal_nan=equal_nan)
        .all()
        .item()
    )


def is_zero(x: Any, /, *, rtol: float | None = None, atol: float | None = None) -> Any:
    """Check if x == 0."""
    return _is_close(x, 0.0, rtol=rtol, atol=atol)


def is_zero_or_finite_and_non_micro(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == 0, or -inf < x < inf and ~isclose(x, 0)."""
    zero = 0.0
    return (x == zero) | is_finite_and_non_zero(x, rtol=rtol, atol=atol)


def is_zero_or_finite_and_non_micro_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == 0, or -inf < x < inf and ~isclose(x, 0), or x == nan."""
    return is_zero_or_finite_and_non_micro(x, rtol=rtol, atol=atol) | isnan(x)


def is_zero_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x > 0 or x == nan."""
    return is_zero(x, rtol=rtol, atol=atol) | isnan(x)


def is_zero_or_non_micro(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == 0 or ~isclose(x, 0)."""
    zero = 0.0
    return (x == zero) | is_non_zero(x, rtol=rtol, atol=atol)


def is_zero_or_non_micro_or_nan(
    x: Any, /, *, rtol: float | None = None, atol: float | None = None
) -> Any:
    """Check if x == 0 or ~isclose(x, 0) or x == nan."""
    return is_zero_or_non_micro(x, rtol=rtol, atol=atol) | isnan(x)


@overload
def maximum(x: float, /) -> float: ...
@overload
def maximum(x0: float, x1: float, /) -> float: ...
@overload
def maximum(x0: float, x1: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: float, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: float, x1: float, x2: float, /) -> float: ...
@overload
def maximum(x0: float, x1: float, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: float, x1: NDArrayF, x2: float, /) -> NDArrayF: ...
@overload
def maximum(x0: float, x1: NDArrayF, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: float, x2: float, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: float, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: NDArrayF, x2: float, /) -> NDArrayF: ...
@overload
def maximum(x0: NDArrayF, x1: NDArrayF, x2: NDArrayF, /) -> NDArrayF: ...
def maximum(*xs: float | NDArrayF) -> float | NDArrayF:
    """Compute the maximum of a number of quantities."""
    return reduce(np.maximum, xs)


@overload
def minimum(x: float, /) -> float: ...
@overload
def minimum(x0: float, x1: float, /) -> float: ...
@overload
def minimum(x0: float, x1: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: float, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: float, x1: float, x2: float, /) -> float: ...
@overload
def minimum(x0: float, x1: float, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: float, x1: NDArrayF, x2: float, /) -> NDArrayF: ...
@overload
def minimum(x0: float, x1: NDArrayF, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: float, x2: float, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: float, x2: NDArrayF, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: NDArrayF, x2: float, /) -> NDArrayF: ...
@overload
def minimum(x0: NDArrayF, x1: NDArrayF, x2: NDArrayF, /) -> NDArrayF: ...
def minimum(*xs: float | NDArrayF) -> float | NDArrayF:
    """Compute the minimum of a number of quantities."""
    return reduce(np.minimum, xs)


def pct_change(
    array: NDArrayF | NDArrayI,
    /,
    *,
    limit: int | None = None,
    n: int = 1,
    axis: int = -1,
) -> NDArrayF:
    """Compute the percentage change in an array."""
    if n == 0:
        raise PctChangeError
    if n > 0:
        filled = ffill(array.astype(float), limit=limit, axis=axis)
        shifted = shift(filled, n=n, axis=axis)
        with errstate(all="ignore"):
            ratio = (filled / shifted) if n >= 0 else (shifted / filled)
        return where(isfinite(array), ratio - 1.0, nan)
    flipped = cast(NDArrayF | NDArrayI, flip(array, axis=axis))
    result = pct_change(flipped, limit=limit, n=-n, axis=axis)
    return flip(result, axis=axis)


@dataclass(kw_only=True)
class PctChangeError(Exception):
    @override
    def __str__(self) -> str:
        return "Shift must be non-zero"


@contextmanager
def redirect_empty_numpy_concatenate() -> Iterator[None]:
    """Redirect to the `EmptyNumpyConcatenateError`."""
    with redirect_error(
        ValueError,
        EmptyNumpyConcatenateError,
        match="need at least one array to concatenate",
    ):
        yield


class EmptyNumpyConcatenateError(Exception): ...


def shift(array: NDArrayF | NDArrayI, /, *, n: int = 1, axis: int = -1) -> NDArrayF:
    """Shift the elements of an array."""
    if n == 0:
        raise ShiftError
    as_float = array.astype(float)
    shifted = roll(as_float, n, axis=axis)
    indexer = list(repeat(slice(None), times=array.ndim))
    indexer[axis] = slice(n) if n >= 0 else slice(n, None)
    shifted[tuple(indexer)] = nan
    return shifted


@dataclass(kw_only=True)
class ShiftError(Exception):
    @override
    def __str__(self) -> str:
        return "Shift must be non-zero"


def shift_bool(
    array: NDArrayB, /, *, n: int = 1, axis: int = -1, fill_value: bool = False
) -> NDArrayB:
    """Shift the elements of a boolean array."""
    shifted = shift(array.astype(float), n=n, axis=axis)
    return fillna(shifted, value=float(fill_value)).astype(bool)


@overload
def year(date: datetime64, /) -> int: ...
@overload
def year(date: NDArrayDD, /) -> NDArrayI: ...
def year(date: datetime64 | NDArrayDD, /) -> int | NDArrayI:
    """Convert a date/array of dates into a year/array of years."""
    years = 1970 + date.astype(datetime64Y).astype(int)
    return years if isinstance(date, ndarray) else years.item()


# annotations - int & predicates
NDArrayINeg = Annotated[NDArrayI, is_negative]
NDArrayINonNeg = Annotated[NDArrayI, is_non_negative]
NDArrayINonPos = Annotated[NDArrayI, is_non_positive]
NDArrayINonZr = Annotated[NDArrayI, is_non_zero]
NDArrayIPos = Annotated[NDArrayI, is_positive]
NDArrayIZr = Annotated[NDArrayI, is_zero]


# annotations - float & predicates
NDArrayFFin = Annotated[NDArrayF, isfinite]
NDArrayFFinInt = Annotated[NDArrayF, is_finite_and_integral]
NDArrayFFinIntNan = Annotated[NDArrayF, is_finite_and_integral_or_nan]
NDArrayFFinNeg = Annotated[NDArrayF, is_finite_and_negative]
NDArrayFFinNegNan = Annotated[NDArrayF, is_finite_and_negative_or_nan]
NDArrayFFinNonNeg = Annotated[NDArrayF, is_finite_and_non_negative]
NDArrayFFinNonNegNan = Annotated[NDArrayF, is_finite_and_non_negative_or_nan]
NDArrayFFinNonPos = Annotated[NDArrayF, is_finite_and_non_positive]
NDArrayFFinNonPosNan = Annotated[NDArrayF, is_finite_and_non_positive_or_nan]
NDArrayFFinNonZr = Annotated[NDArrayF, is_finite_and_non_zero]
NDArrayFFinNonZrNan = Annotated[NDArrayF, is_finite_and_non_zero_or_nan]
NDArrayFFinPos = Annotated[NDArrayF, is_finite_and_positive]
NDArrayFFinPosNan = Annotated[NDArrayF, is_finite_and_positive_or_nan]
NDArrayFFinNan = Annotated[NDArrayF, is_finite_or_nan]
NDArrayFInt = Annotated[NDArrayF, is_integral]
NDArrayFIntNan = Annotated[NDArrayF, is_integral_or_nan]
NDArrayFNeg = Annotated[NDArrayF, is_negative]
NDArrayFNegNan = Annotated[NDArrayF, is_negative_or_nan]
NDArrayFNonNeg = Annotated[NDArrayF, is_non_negative]
NDArrayFNonNegNan = Annotated[NDArrayF, is_non_negative_or_nan]
NDArrayFNonPos = Annotated[NDArrayF, is_non_positive]
NDArrayFNonPosNan = Annotated[NDArrayF, is_non_positive_or_nan]
NDArrayFNonZr = Annotated[NDArrayF, is_non_zero]
NDArrayFNonZrNan = Annotated[NDArrayF, is_non_zero_or_nan]
NDArrayFPos = Annotated[NDArrayF, is_positive]
NDArrayFPosNan = Annotated[NDArrayF, is_positive_or_nan]
NDArrayFZr = Annotated[NDArrayF, is_zero]
NDArrayFZrFinNonMic = Annotated[NDArrayF, is_zero_or_finite_and_non_micro]
NDArrayFZrFinNonMicNan = Annotated[NDArrayF, is_zero_or_finite_and_non_micro_or_nan]
NDArrayFZrNan = Annotated[NDArrayF, is_zero_or_nan]
NDArrayFZrNonMic = Annotated[NDArrayF, is_zero_or_non_micro]
NDArrayFZrNonMicNan = Annotated[NDArrayF, is_zero_or_non_micro_or_nan]


# annotations - int, ndim & predicate
NDArrayI0Neg = Annotated[NDArrayI0, is_negative]
NDArrayI0NonNeg = Annotated[NDArrayI0, is_non_negative]
NDArrayI0NonPos = Annotated[NDArrayI0, is_non_positive]
NDArrayI0NonZr = Annotated[NDArrayI0, is_non_zero]
NDArrayI0Pos = Annotated[NDArrayI0, is_positive]
NDArrayI0Zr = Annotated[NDArrayI0, is_zero]

NDArrayI1Neg = Annotated[NDArrayI1, is_negative]
NDArrayI1NonNeg = Annotated[NDArrayI1, is_non_negative]
NDArrayI1NonPos = Annotated[NDArrayI1, is_non_positive]
NDArrayI1NonZr = Annotated[NDArrayI1, is_non_zero]
NDArrayI1Pos = Annotated[NDArrayI1, is_positive]
NDArrayI1Zr = Annotated[NDArrayI1, is_zero]

NDArrayI2Neg = Annotated[NDArrayI2, is_negative]
NDArrayI2NonNeg = Annotated[NDArrayI2, is_non_negative]
NDArrayI2NonPos = Annotated[NDArrayI2, is_non_positive]
NDArrayI2NonZr = Annotated[NDArrayI2, is_non_zero]
NDArrayI2Pos = Annotated[NDArrayI2, is_positive]
NDArrayI2Zr = Annotated[NDArrayI2, is_zero]

NDArrayI3Neg = Annotated[NDArrayI1, is_negative]
NDArrayI3NonNeg = Annotated[NDArrayI3, is_non_negative]
NDArrayI3NonPos = Annotated[NDArrayI3, is_non_positive]
NDArrayI3NonZr = Annotated[NDArrayI3, is_non_zero]
NDArrayI3Pos = Annotated[NDArrayI3, is_positive]
NDArrayI3Zr = Annotated[NDArrayI3, is_zero]


# annotations - float, ndim & predicate
NDArrayF0Fin = Annotated[NDArrayF0, isfinite]
NDArrayF0FinInt = Annotated[NDArrayF0, is_finite_and_integral]
NDArrayF0FinIntNan = Annotated[NDArrayF0, is_finite_and_integral_or_nan]
NDArrayF0FinNeg = Annotated[NDArrayF0, is_finite_and_negative]
NDArrayF0FinNegNan = Annotated[NDArrayF0, is_finite_and_negative_or_nan]
NDArrayF0FinNonNeg = Annotated[NDArrayF0, is_finite_and_non_negative]
NDArrayF0FinNonNegNan = Annotated[NDArrayF0, is_finite_and_non_negative_or_nan]
NDArrayF0FinNonPos = Annotated[NDArrayF0, is_finite_and_non_positive]
NDArrayF0FinNonPosNan = Annotated[NDArrayF0, is_finite_and_non_positive_or_nan]
NDArrayF0FinNonZr = Annotated[NDArrayF0, is_finite_and_non_zero]
NDArrayF0FinNonZrNan = Annotated[NDArrayF0, is_finite_and_non_zero_or_nan]
NDArrayF0FinPos = Annotated[NDArrayF0, is_finite_and_positive]
NDArrayF0FinPosNan = Annotated[NDArrayF0, is_finite_and_positive_or_nan]
NDArrayF0FinNan = Annotated[NDArrayF0, is_finite_or_nan]
NDArrayF0Int = Annotated[NDArrayF0, is_integral]
NDArrayF0IntNan = Annotated[NDArrayF0, is_integral_or_nan]
NDArrayF0Neg = Annotated[NDArrayF0, is_negative]
NDArrayF0NegNan = Annotated[NDArrayF0, is_negative_or_nan]
NDArrayF0NonNeg = Annotated[NDArrayF0, is_non_negative]
NDArrayF0NonNegNan = Annotated[NDArrayF0, is_non_negative_or_nan]
NDArrayF0NonPos = Annotated[NDArrayF0, is_non_positive]
NDArrayF0NonPosNan = Annotated[NDArrayF0, is_non_positive_or_nan]
NDArrayF0NonZr = Annotated[NDArrayF0, is_non_zero]
NDArrayF0NonZrNan = Annotated[NDArrayF0, is_non_zero_or_nan]
NDArrayF0Pos = Annotated[NDArrayF0, is_positive]
NDArrayF0PosNan = Annotated[NDArrayF0, is_positive_or_nan]
NDArrayF0Zr = Annotated[NDArrayF0, is_zero]
NDArrayF0ZrFinNonMic = Annotated[NDArrayF0, is_zero_or_finite_and_non_micro]
NDArrayF0ZrFinNonMicNan = Annotated[NDArrayF0, is_zero_or_finite_and_non_micro_or_nan]
NDArrayF0ZrNan = Annotated[NDArrayF0, is_zero_or_nan]
NDArrayF0ZrNonMic = Annotated[NDArrayF0, is_zero_or_non_micro]
NDArrayF0ZrNonMicNan = Annotated[NDArrayF0, is_zero_or_non_micro_or_nan]

NDArrayF1Fin = Annotated[NDArrayF1, isfinite]
NDArrayF1FinInt = Annotated[NDArrayF1, is_finite_and_integral]
NDArrayF1FinIntNan = Annotated[NDArrayF1, is_finite_and_integral_or_nan]
NDArrayF1FinNeg = Annotated[NDArrayF1, is_finite_and_negative]
NDArrayF1FinNegNan = Annotated[NDArrayF1, is_finite_and_negative_or_nan]
NDArrayF1FinNonNeg = Annotated[NDArrayF1, is_finite_and_non_negative]
NDArrayF1FinNonNegNan = Annotated[NDArrayF1, is_finite_and_non_negative_or_nan]
NDArrayF1FinNonPos = Annotated[NDArrayF1, is_finite_and_non_positive]
NDArrayF1FinNonPosNan = Annotated[NDArrayF1, is_finite_and_non_positive_or_nan]
NDArrayF1FinNonZr = Annotated[NDArrayF1, is_finite_and_non_zero]
NDArrayF1FinNonZrNan = Annotated[NDArrayF1, is_finite_and_non_zero_or_nan]
NDArrayF1FinPos = Annotated[NDArrayF1, is_finite_and_positive]
NDArrayF1FinPosNan = Annotated[NDArrayF1, is_finite_and_positive_or_nan]
NDArrayF1FinNan = Annotated[NDArrayF1, is_finite_or_nan]
NDArrayF1Int = Annotated[NDArrayF1, is_integral]
NDArrayF1IntNan = Annotated[NDArrayF1, is_integral_or_nan]
NDArrayF1Neg = Annotated[NDArrayF1, is_negative]
NDArrayF1NegNan = Annotated[NDArrayF1, is_negative_or_nan]
NDArrayF1NonNeg = Annotated[NDArrayF1, is_non_negative]
NDArrayF1NonNegNan = Annotated[NDArrayF1, is_non_negative_or_nan]
NDArrayF1NonPos = Annotated[NDArrayF1, is_non_positive]
NDArrayF1NonPosNan = Annotated[NDArrayF1, is_non_positive_or_nan]
NDArrayF1NonZr = Annotated[NDArrayF1, is_non_zero]
NDArrayF1NonZrNan = Annotated[NDArrayF1, is_non_zero_or_nan]
NDArrayF1Pos = Annotated[NDArrayF1, is_positive]
NDArrayF1PosNan = Annotated[NDArrayF1, is_positive_or_nan]
NDArrayF1Zr = Annotated[NDArrayF1, is_zero]
NDArrayF1ZrFinNonMic = Annotated[NDArrayF1, is_zero_or_finite_and_non_micro]
NDArrayF1ZrFinNonMicNan = Annotated[NDArrayF1, is_zero_or_finite_and_non_micro_or_nan]
NDArrayF1ZrNan = Annotated[NDArrayF1, is_zero_or_nan]
NDArrayF1ZrNonMic = Annotated[NDArrayF1, is_zero_or_non_micro]
NDArrayF1ZrNonMicNan = Annotated[NDArrayF1, is_zero_or_non_micro_or_nan]

NDArrayF2Fin = Annotated[NDArrayF2, isfinite]
NDArrayF2FinInt = Annotated[NDArrayF2, is_finite_and_integral]
NDArrayF2FinIntNan = Annotated[NDArrayF2, is_finite_and_integral_or_nan]
NDArrayF2FinNeg = Annotated[NDArrayF2, is_finite_and_negative]
NDArrayF2FinNegNan = Annotated[NDArrayF2, is_finite_and_negative_or_nan]
NDArrayF2FinNonNeg = Annotated[NDArrayF2, is_finite_and_non_negative]
NDArrayF2FinNonNegNan = Annotated[NDArrayF2, is_finite_and_non_negative_or_nan]
NDArrayF2FinNonPos = Annotated[NDArrayF2, is_finite_and_non_positive]
NDArrayF2FinNonPosNan = Annotated[NDArrayF2, is_finite_and_non_positive_or_nan]
NDArrayF2FinNonZr = Annotated[NDArrayF2, is_finite_and_non_zero]
NDArrayF2FinNonZrNan = Annotated[NDArrayF2, is_finite_and_non_zero_or_nan]
NDArrayF2FinPos = Annotated[NDArrayF2, is_finite_and_positive]
NDArrayF2FinPosNan = Annotated[NDArrayF2, is_finite_and_positive_or_nan]
NDArrayF2FinNan = Annotated[NDArrayF2, is_finite_or_nan]
NDArrayF2Int = Annotated[NDArrayF2, is_integral]
NDArrayF2IntNan = Annotated[NDArrayF2, is_integral_or_nan]
NDArrayF2Neg = Annotated[NDArrayF2, is_negative]
NDArrayF2NegNan = Annotated[NDArrayF2, is_negative_or_nan]
NDArrayF2NonNeg = Annotated[NDArrayF2, is_non_negative]
NDArrayF2NonNegNan = Annotated[NDArrayF2, is_non_negative_or_nan]
NDArrayF2NonPos = Annotated[NDArrayF2, is_non_positive]
NDArrayF2NonPosNan = Annotated[NDArrayF2, is_non_positive_or_nan]
NDArrayF2NonZr = Annotated[NDArrayF2, is_non_zero]
NDArrayF2NonZrNan = Annotated[NDArrayF2, is_non_zero_or_nan]
NDArrayF2Pos = Annotated[NDArrayF2, is_positive]
NDArrayF2PosNan = Annotated[NDArrayF2, is_positive_or_nan]
NDArrayF2Zr = Annotated[NDArrayF2, is_zero]
NDArrayF2ZrFinNonMic = Annotated[NDArrayF2, is_zero_or_finite_and_non_micro]
NDArrayF2ZrFinNonMicNan = Annotated[NDArrayF2, is_zero_or_finite_and_non_micro_or_nan]
NDArrayF2ZrNan = Annotated[NDArrayF2, is_zero_or_nan]
NDArrayF2ZrNonMic = Annotated[NDArrayF2, is_zero_or_non_micro]
NDArrayF2ZrNonMicNan = Annotated[NDArrayF2, is_zero_or_non_micro_or_nan]

NDArrayF3Fin = Annotated[NDArrayF3, isfinite]
NDArrayF3FinInt = Annotated[NDArrayF3, is_finite_and_integral]
NDArrayF3FinIntNan = Annotated[NDArrayF3, is_finite_and_integral_or_nan]
NDArrayF3FinNeg = Annotated[NDArrayF3, is_finite_and_negative]
NDArrayF3FinNegNan = Annotated[NDArrayF3, is_finite_and_negative_or_nan]
NDArrayF3FinNonNeg = Annotated[NDArrayF3, is_finite_and_non_negative]
NDArrayF3FinNonNegNan = Annotated[NDArrayF3, is_finite_and_non_negative_or_nan]
NDArrayF3FinNonPos = Annotated[NDArrayF3, is_finite_and_non_positive]
NDArrayF3FinNonPosNan = Annotated[NDArrayF3, is_finite_and_non_positive_or_nan]
NDArrayF3FinNonZr = Annotated[NDArrayF3, is_finite_and_non_zero]
NDArrayF3FinNonZrNan = Annotated[NDArrayF3, is_finite_and_non_zero_or_nan]
NDArrayF3FinPos = Annotated[NDArrayF3, is_finite_and_positive]
NDArrayF3FinPosNan = Annotated[NDArrayF3, is_finite_and_positive_or_nan]
NDArrayF3FinNan = Annotated[NDArrayF3, is_finite_or_nan]
NDArrayF3Int = Annotated[NDArrayF3, is_integral]
NDArrayF3IntNan = Annotated[NDArrayF3, is_integral_or_nan]
NDArrayF3Neg = Annotated[NDArrayF3, is_negative]
NDArrayF3NegNan = Annotated[NDArrayF3, is_negative_or_nan]
NDArrayF3NonNeg = Annotated[NDArrayF3, is_non_negative]
NDArrayF3NonNegNan = Annotated[NDArrayF3, is_non_negative_or_nan]
NDArrayF3NonPos = Annotated[NDArrayF3, is_non_positive]
NDArrayF3NonPosNan = Annotated[NDArrayF3, is_non_positive_or_nan]
NDArrayF3NonZr = Annotated[NDArrayF3, is_non_zero]
NDArrayF3NonZrNan = Annotated[NDArrayF3, is_non_zero_or_nan]
NDArrayF3Pos = Annotated[NDArrayF3, is_positive]
NDArrayF3PosNan = Annotated[NDArrayF3, is_positive_or_nan]
NDArrayF3Zr = Annotated[NDArrayF3, is_zero]
NDArrayF3ZrFinNonMic = Annotated[NDArrayF3, is_zero_or_finite_and_non_micro]
NDArrayF3ZrFinNonMicNan = Annotated[NDArrayF3, is_zero_or_finite_and_non_micro_or_nan]
NDArrayF3ZrNan = Annotated[NDArrayF3, is_zero_or_nan]
NDArrayF3ZrNonMic = Annotated[NDArrayF3, is_zero_or_non_micro]
NDArrayF3ZrNonMicNan = Annotated[NDArrayF3, is_zero_or_non_micro_or_nan]


__all__ = [
    "DATETIME_MAX_AS_DATETIME64",
    "DATETIME_MAX_AS_INT",
    "DATETIME_MIN_AS_DATETIME64",
    "DATETIME_MIN_AS_INT",
    "DATE_MAX_AS_DATETIME64",
    "DATE_MAX_AS_INT",
    "DATE_MIN_AS_DATETIME64",
    "DATE_MIN_AS_INT",
    "DEFAULT_RNG",
    "AsIntError",
    "DateTime64ToDateError",
    "DateTime64ToDateTimeError",
    "Datetime64Kind",
    "Datetime64Unit",
    "EmptyNumpyConcatenateError",
    "FlatN0EmptyError",
    "FlatN0Error",
    "FlatN0MultipleError",
    "GetFillValueError",
    "NDArray0",
    "NDArray1",
    "NDArray2",
    "NDArray3",
    "NDArrayA",
    "NDArrayB",
    "NDArrayB0",
    "NDArrayB1",
    "NDArrayB2",
    "NDArrayB3",
    "NDArrayD",
    "NDArrayD0",
    "NDArrayD1",
    "NDArrayD2",
    "NDArrayD3",
    "NDArrayDD",
    "NDArrayDD0",
    "NDArrayDD1",
    "NDArrayDD2",
    "NDArrayDD3",
    "NDArrayDM",
    "NDArrayDM0",
    "NDArrayDM1",
    "NDArrayDM2",
    "NDArrayDM3",
    "NDArrayDW",
    "NDArrayDW0",
    "NDArrayDW1",
    "NDArrayDW2",
    "NDArrayDW3",
    "NDArrayDY",
    "NDArrayDY0",
    "NDArrayDY1",
    "NDArrayDY2",
    "NDArrayDY3",
    "NDArrayDas",
    "NDArrayDas0",
    "NDArrayDas1",
    "NDArrayDas2",
    "NDArrayDas3",
    "NDArrayDfs",
    "NDArrayDfs0",
    "NDArrayDfs1",
    "NDArrayDfs2",
    "NDArrayDfs3",
    "NDArrayDh",
    "NDArrayDh0",
    "NDArrayDh1",
    "NDArrayDh2",
    "NDArrayDh3",
    "NDArrayDm",
    "NDArrayDm0",
    "NDArrayDm1",
    "NDArrayDm2",
    "NDArrayDm3",
    "NDArrayDms",
    "NDArrayDms0",
    "NDArrayDms1",
    "NDArrayDms2",
    "NDArrayDms3",
    "NDArrayDns",
    "NDArrayDns0",
    "NDArrayDns1",
    "NDArrayDns2",
    "NDArrayDns3",
    "NDArrayDps",
    "NDArrayDps0",
    "NDArrayDps1",
    "NDArrayDps2",
    "NDArrayDps3",
    "NDArrayDs",
    "NDArrayDs0",
    "NDArrayDs1",
    "NDArrayDs2",
    "NDArrayDs3",
    "NDArrayDus",
    "NDArrayDus0",
    "NDArrayDus1",
    "NDArrayDus2",
    "NDArrayDus3",
    "NDArrayF",
    "NDArrayF0",
    "NDArrayF0Fin",
    "NDArrayF0FinInt",
    "NDArrayF0FinIntNan",
    "NDArrayF0FinNan",
    "NDArrayF0FinNeg",
    "NDArrayF0FinNegNan",
    "NDArrayF0FinNonNeg",
    "NDArrayF0FinNonNegNan",
    "NDArrayF0FinNonPos",
    "NDArrayF0FinNonPosNan",
    "NDArrayF0FinNonZr",
    "NDArrayF0FinNonZrNan",
    "NDArrayF0FinPos",
    "NDArrayF0FinPosNan",
    "NDArrayF0Int",
    "NDArrayF0IntNan",
    "NDArrayF0Neg",
    "NDArrayF0NegNan",
    "NDArrayF0NonNeg",
    "NDArrayF0NonNegNan",
    "NDArrayF0NonPos",
    "NDArrayF0NonPosNan",
    "NDArrayF0NonZr",
    "NDArrayF0NonZrNan",
    "NDArrayF0Pos",
    "NDArrayF0PosNan",
    "NDArrayF0Zr",
    "NDArrayF0ZrFinNonMic",
    "NDArrayF0ZrFinNonMicNan",
    "NDArrayF0ZrNan",
    "NDArrayF0ZrNonMic",
    "NDArrayF0ZrNonMicNan",
    "NDArrayF1",
    "NDArrayF1Fin",
    "NDArrayF1FinInt",
    "NDArrayF1FinIntNan",
    "NDArrayF1FinNan",
    "NDArrayF1FinNeg",
    "NDArrayF1FinNegNan",
    "NDArrayF1FinNonNeg",
    "NDArrayF1FinNonNegNan",
    "NDArrayF1FinNonPos",
    "NDArrayF1FinNonPosNan",
    "NDArrayF1FinNonZr",
    "NDArrayF1FinNonZrNan",
    "NDArrayF1FinPos",
    "NDArrayF1FinPosNan",
    "NDArrayF1Int",
    "NDArrayF1IntNan",
    "NDArrayF1Neg",
    "NDArrayF1NegNan",
    "NDArrayF1NonNeg",
    "NDArrayF1NonNegNan",
    "NDArrayF1NonPos",
    "NDArrayF1NonPosNan",
    "NDArrayF1NonZr",
    "NDArrayF1NonZrNan",
    "NDArrayF1Pos",
    "NDArrayF1PosNan",
    "NDArrayF1Zr",
    "NDArrayF1ZrFinNonMic",
    "NDArrayF1ZrFinNonMicNan",
    "NDArrayF1ZrNan",
    "NDArrayF1ZrNonMic",
    "NDArrayF1ZrNonMicNan",
    "NDArrayF2",
    "NDArrayF2Fin",
    "NDArrayF2FinInt",
    "NDArrayF2FinIntNan",
    "NDArrayF2FinNan",
    "NDArrayF2FinNeg",
    "NDArrayF2FinNegNan",
    "NDArrayF2FinNonNeg",
    "NDArrayF2FinNonNegNan",
    "NDArrayF2FinNonPos",
    "NDArrayF2FinNonPosNan",
    "NDArrayF2FinNonZr",
    "NDArrayF2FinNonZrNan",
    "NDArrayF2FinPos",
    "NDArrayF2FinPosNan",
    "NDArrayF2Int",
    "NDArrayF2IntNan",
    "NDArrayF2Neg",
    "NDArrayF2NegNan",
    "NDArrayF2NonNeg",
    "NDArrayF2NonNegNan",
    "NDArrayF2NonPos",
    "NDArrayF2NonPosNan",
    "NDArrayF2NonZr",
    "NDArrayF2NonZrNan",
    "NDArrayF2Pos",
    "NDArrayF2PosNan",
    "NDArrayF2Zr",
    "NDArrayF2ZrFinNonMic",
    "NDArrayF2ZrFinNonMicNan",
    "NDArrayF2ZrNan",
    "NDArrayF2ZrNonMic",
    "NDArrayF2ZrNonMicNan",
    "NDArrayF3",
    "NDArrayF3Fin",
    "NDArrayF3FinInt",
    "NDArrayF3FinIntNan",
    "NDArrayF3FinNan",
    "NDArrayF3FinNeg",
    "NDArrayF3FinNegNan",
    "NDArrayF3FinNonNeg",
    "NDArrayF3FinNonNegNan",
    "NDArrayF3FinNonPos",
    "NDArrayF3FinNonPosNan",
    "NDArrayF3FinNonZr",
    "NDArrayF3FinNonZrNan",
    "NDArrayF3FinPos",
    "NDArrayF3FinPosNan",
    "NDArrayF3Int",
    "NDArrayF3IntNan",
    "NDArrayF3Neg",
    "NDArrayF3NegNan",
    "NDArrayF3NonNeg",
    "NDArrayF3NonNegNan",
    "NDArrayF3NonPos",
    "NDArrayF3NonPosNan",
    "NDArrayF3NonZr",
    "NDArrayF3NonZrNan",
    "NDArrayF3Pos",
    "NDArrayF3PosNan",
    "NDArrayF3Zr",
    "NDArrayF3ZrFinNonMic",
    "NDArrayF3ZrFinNonMicNan",
    "NDArrayF3ZrNan",
    "NDArrayF3ZrNonMic",
    "NDArrayF3ZrNonMicNan",
    "NDArrayFFin",
    "NDArrayFFinInt",
    "NDArrayFFinIntNan",
    "NDArrayFFinNan",
    "NDArrayFFinNeg",
    "NDArrayFFinNegNan",
    "NDArrayFFinNonNeg",
    "NDArrayFFinNonNegNan",
    "NDArrayFFinNonPos",
    "NDArrayFFinNonPosNan",
    "NDArrayFFinNonZr",
    "NDArrayFFinNonZrNan",
    "NDArrayFFinPos",
    "NDArrayFFinPosNan",
    "NDArrayFInt",
    "NDArrayFIntNan",
    "NDArrayFNeg",
    "NDArrayFNegNan",
    "NDArrayFNonNeg",
    "NDArrayFNonNegNan",
    "NDArrayFNonPos",
    "NDArrayFNonPosNan",
    "NDArrayFNonZr",
    "NDArrayFNonZrNan",
    "NDArrayFPos",
    "NDArrayFPosNan",
    "NDArrayFZr",
    "NDArrayFZrFinNonMic",
    "NDArrayFZrFinNonMicNan",
    "NDArrayFZrNan",
    "NDArrayFZrNonMic",
    "NDArrayFZrNonMicNan",
    "NDArrayI",
    "NDArrayI0",
    "NDArrayI0Neg",
    "NDArrayI0NonNeg",
    "NDArrayI0NonPos",
    "NDArrayI0NonZr",
    "NDArrayI0Pos",
    "NDArrayI0Zr",
    "NDArrayI1",
    "NDArrayI1Neg",
    "NDArrayI1NonNeg",
    "NDArrayI1NonPos",
    "NDArrayI1NonZr",
    "NDArrayI1Pos",
    "NDArrayI1Zr",
    "NDArrayI2",
    "NDArrayI2Neg",
    "NDArrayI2NonNeg",
    "NDArrayI2NonPos",
    "NDArrayI2NonZr",
    "NDArrayI2Pos",
    "NDArrayI2Zr",
    "NDArrayI3",
    "NDArrayI3Neg",
    "NDArrayI3NonNeg",
    "NDArrayI3NonPos",
    "NDArrayI3NonZr",
    "NDArrayI3Pos",
    "NDArrayI3Zr",
    "NDArrayINeg",
    "NDArrayINonNeg",
    "NDArrayINonPos",
    "NDArrayINonZr",
    "NDArrayIPos",
    "NDArrayIZr",
    "NDArrayO",
    "NDArrayO0",
    "NDArrayO1",
    "NDArrayO2",
    "NDArrayO3",
    "PctChangeError",
    "ShiftError",
    "array_indexer",
    "as_int",
    "date_to_datetime64",
    "datetime64D",
    "datetime64M",
    "datetime64W",
    "datetime64Y",
    "datetime64_to_date",
    "datetime64_to_datetime",
    "datetime64_to_int",
    "datetime64as",
    "datetime64fs",
    "datetime64h",
    "datetime64m",
    "datetime64ms",
    "datetime64ns",
    "datetime64ps",
    "datetime64s",
    "datetime64us",
    "datetime_to_datetime64",
    "discretize",
    "ewma",
    "exp_moving_sum",
    "ffill",
    "ffill_non_nan_slices",
    "fillna",
    "flatn0",
    "get_fill_value",
    "has_dtype",
    "is_at_least",
    "is_at_least_or_nan",
    "is_at_most",
    "is_at_most_or_nan",
    "is_between",
    "is_between_or_nan",
    "is_empty",
    "is_finite_and_integral",
    "is_finite_and_integral_or_nan",
    "is_finite_and_negative",
    "is_finite_and_negative_or_nan",
    "is_finite_and_non_negative",
    "is_finite_and_non_negative_or_nan",
    "is_finite_and_non_positive",
    "is_finite_and_non_positive_or_nan",
    "is_finite_and_non_zero",
    "is_finite_and_non_zero_or_nan",
    "is_finite_and_positive",
    "is_finite_and_positive_or_nan",
    "is_finite_or_nan",
    "is_greater_than",
    "is_greater_than_or_nan",
    "is_integral",
    "is_integral_or_nan",
    "is_less_than",
    "is_less_than_or_nan",
    "is_negative",
    "is_negative_or_nan",
    "is_non_empty",
    "is_non_negative",
    "is_non_negative_or_nan",
    "is_non_positive",
    "is_non_positive_or_nan",
    "is_non_singular",
    "is_non_zero",
    "is_non_zero_or_nan",
    "is_positive",
    "is_positive_or_nan",
    "is_positive_semidefinite",
    "is_symmetric",
    "is_zero",
    "is_zero_or_finite_and_non_micro",
    "is_zero_or_finite_and_non_micro_or_nan",
    "is_zero_or_nan",
    "is_zero_or_non_micro",
    "is_zero_or_non_micro_or_nan",
    "maximum",
    "minimum",
    "pct_change",
    "redirect_empty_numpy_concatenate",
    "shift",
    "shift_bool",
    "year",
]



from __future__ import annotations

import datetime as dt
from dataclasses import dataclass
from numbers import Number
from operator import eq, ge, gt, le, lt, ne
from timeit import default_timer
from typing import TYPE_CHECKING, Any

from typing_extensions import Self, override

from utilities.types import EnsureClassError, ensure_class, get_class_name

if TYPE_CHECKING:
    from collections.abc import Callable


class Timer:
    """Context manager for timing blocks of code."""

    def __init__(self) -> None:
        super().__init__()
        self._start = default_timer()
        self._end: float | None = None

    def __enter__(self) -> Self:
        self._start = default_timer()
        return self

    def __exit__(self, *_: object) -> bool:
        self._end = default_timer()
        return False

    def __float__(self) -> float:
        end_use = default_timer() if (end := self._end) is None else end
        return end_use - self._start

    @override
    def __repr__(self) -> str:
        return str(self.timedelta)

    @override
    def __str__(self) -> str:
        return str(self.timedelta)

    @override
    def __eq__(self, other: object) -> bool:
        return self._compare(other, eq)

    def __ge__(self, other: Any) -> bool:
        return self._compare(other, ge)

    def __gt__(self, other: Any) -> bool:
        return self._compare(other, gt)

    def __le__(self, other: Any) -> bool:
        return self._compare(other, le)

    def __lt__(self, other: Any) -> bool:
        return self._compare(other, lt)

    @override
    def __ne__(self, other: object) -> bool:
        return self._compare(other, ne)

    @property
    def timedelta(self) -> dt.timedelta:
        """The elapsed time, as a `timedelta` object."""
        return dt.timedelta(seconds=float(self))

    def _compare(self, other: Any, op: Callable[[Any, Any], bool], /) -> bool:
        try:
            right = ensure_class(other, (Number, Timer, dt.timedelta))
        except EnsureClassError:
            raise TimerError(obj=other) from None
        left = float(self) if isinstance(right, Number | Timer) else self.timedelta
        return op(left, right)


@dataclass(kw_only=True)
class TimerError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return (
            "Timer must be compared to a number, Timer, or timedelta; got "
            f"{get_class_name(self.obj)} instead"
        )


__all__ = ["Timer", "TimerError"]



from __future__ import annotations

import datetime as dt
import re
from re import DOTALL
from typing import TYPE_CHECKING, Any, Literal, cast

from hypothesis import assume, given
from hypothesis.strategies import none
from numpy import array, nan
from numpy.testing import assert_equal
from pandas import (
    NA,
    DataFrame,
    Index,
    NaT,
    RangeIndex,
    Series,
    Timestamp,
    concat,
    to_datetime,
)
from pandas.testing import assert_frame_equal, assert_index_equal, assert_series_equal
from pytest import mark, param, raises

from utilities.datetime import TODAY_UTC, UTC
from utilities.hypothesis import text_ascii, timestamps
from utilities.numpy import datetime64ns
from utilities.pandas import (
    TIMESTAMP_MAX_AS_DATE,
    TIMESTAMP_MAX_AS_DATETIME,
    TIMESTAMP_MIN_AS_DATE,
    TIMESTAMP_MIN_AS_DATETIME,
    AssignBeforeOrAfterError,
    AssignBetweenIndexError,
    AssignBetweenIndicesError,
    CheckIndexError,
    CheckPandasDataFrameError,
    CheckRangeIndexError,
    EmptyPandasConcatError,
    Int64,
    ReindexToSetError,
    ReindexToSubSetError,
    ReindexToSuperSetError,
    SeriesA,
    SeriesMinMaxError,
    TimestampToDateTimeError,
    UnionIndexesError,
    assign_after,
    assign_before,
    assign_between,
    boolean,
    check_index,
    check_pandas_dataframe,
    check_range_index,
    redirect_empty_pandas_concat,
    reindex_to_set,
    reindex_to_subset,
    reindex_to_superset,
    series_max,
    series_min,
    string,
    timestamp_to_date,
    timestamp_to_datetime,
    to_numpy,
    union_indexes,
)

if TYPE_CHECKING:
    from collections.abc import Callable


class TestAssignBeforeOrAfter:
    def test_after(self) -> None:
        df = DataFrame([[0, 1, 2]], columns=Index(["a", "b", "c"]))
        result = assign_after(df, "b", Series(9, name="z"))
        expected = DataFrame([[0, 1, 9, 2]], columns=Index(["a", "b", "z", "c"]))
        assert_frame_equal(result, expected)

    def test_before(self) -> None:
        df = DataFrame([[0, 1, 2]], columns=Index(["a", "b", "c"]))
        result = assign_before(df, "b", Series(9, name="z"))
        expected = DataFrame([[0, 9, 1, 2]], columns=Index(["a", "z", "b", "c"]))
        assert_frame_equal(result, expected)

    @mark.parametrize("mode", [param("before"), param("after")])
    def test_error(self, *, mode: Literal["before", "after"]) -> None:
        df = DataFrame([[0, 1, 2]], columns=Index(["a", "b", "c"]))
        match mode:
            case "before":
                func = assign_before
            case "after":
                func = assign_after
        with raises(
            AssignBeforeOrAfterError,
            match="DataFrame must contain exactly one column named 'd'",
        ):
            _ = func(df, "d", Series(9, name="z"))


class TestAssignBetween:
    def test_main(self) -> None:
        df = DataFrame([[0, 1, 2, 3]], columns=Index(["a", "b", "c", "d"]))
        result = assign_between(df, "b", "c", Series(9, name="z"))
        expected = DataFrame(
            [[0, 1, 9, 2, 3]], columns=Index(["a", "b", "z", "c", "d"])
        )
        assert_frame_equal(result, expected)

    def test_error_index(self) -> None:
        df = DataFrame([[0, 1, 2]], columns=Index(["a", "b", "c"]))
        with raises(
            AssignBetweenIndexError,
            match="DataFrame must contain exactly one column named 'd'",
        ):
            _ = assign_between(df, "d", "e", Series(9, name="z"))

    def test_error_indices(self) -> None:
        df = DataFrame([[0, 1, 2]], columns=Index(["a", "b", "c"]))
        with raises(
            AssignBetweenIndicesError,
            match="DataFrame must specify consecutive indices; got 0 and 2",
        ):
            _ = assign_between(df, "a", "c", Series(9, name="z"))


class TestCheckIndex:
    def test_main(self) -> None:
        check_index(RangeIndex(1))

    def test_length_pass(self) -> None:
        check_index(RangeIndex(1), length=1)

    def test_length_error(self) -> None:
        with raises(
            CheckIndexError, match=r"Index .* must satisfy the length requirements\."
        ):
            check_index(RangeIndex(1), length=2)

    def test_min_length_pass(self) -> None:
        check_index(RangeIndex(2), min_length=1)

    def test_min_length_error(self) -> None:
        with raises(
            CheckIndexError, match=r"Index .* must satisfy the length requirements\."
        ):
            check_index(RangeIndex(0), min_length=1)

    def test_max_length_pass(self) -> None:
        check_index(RangeIndex(0), max_length=1)

    def test_max_length_error(self) -> None:
        with raises(
            CheckIndexError, match=r"Index .* must satisfy the length requirements\."
        ):
            check_index(RangeIndex(2), max_length=1)

    def test_name_pass(self) -> None:
        check_index(RangeIndex(0), name=None)

    def test_name_error(self) -> None:
        with raises(
            CheckIndexError, match=r"Index .* must satisfy the name requirement\."
        ):
            check_index(RangeIndex(0), name="name")

    def test_sorted_pass(self) -> None:
        check_index(Index(["A", "B"]), sorted=True)

    def test_sorted_error(self) -> None:
        with raises(CheckIndexError, match=r"Index .* must be sorted\."):
            check_index(Index(["B", "A"]), sorted=True)

    def test_unique_pass(self) -> None:
        check_index(Index(["A", "B"]), unique=True)

    def test_unique_error(self) -> None:
        with raises(CheckIndexError, match=r"Index .* must be unique\."):
            check_index(Index(["A", "A"]), unique=True)


class TestCheckPandasDataFrame:
    def test_main(self) -> None:
        check_pandas_dataframe(DataFrame())

    def test_columns_pass(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(0), columns=Index([]))
        check_pandas_dataframe(df, columns=[])

    def test_columns_error(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(0), columns=Index(["value"]))
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must have columns .*; got .*\n\n.*\.", flags=DOTALL
            ),
        ):
            check_pandas_dataframe(df, columns=["other"])

    def test_dtypes_pass(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(0), columns=Index([]))
        check_pandas_dataframe(df, dtypes={})

    def test_dtypes_error_set_of_columns(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(0), columns=Index([]))
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must have dtypes .*; got .*\n\n.*\.", flags=DOTALL
            ),
        ):
            check_pandas_dataframe(df, dtypes={"value": int})

    def test_dtypes_error_order_of_columns(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(0), columns=Index(["a", "b"]))
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must have dtypes .*; got .*\n\n.*\.", flags=DOTALL
            ),
        ):
            check_pandas_dataframe(df, dtypes={"b": float, "a": float})

    def test_length_pass(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(1), columns=Index(["value"]))
        check_pandas_dataframe(df, length=1)

    def test_length_error(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(1), columns=Index(["value"]))
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must satisfy the length requirements; got .*\n\n.*\.",
                flags=DOTALL,
            ),
        ):
            check_pandas_dataframe(df, length=2)

    def test_min_length_pass(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(2), columns=Index(["value"]))
        check_pandas_dataframe(df, min_length=1)

    def test_min_length_error(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(0), columns=Index(["value"]))
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must satisfy the length requirements; got .*\n\n.*\.",
                flags=DOTALL,
            ),
        ):
            check_pandas_dataframe(df, min_length=1)

    def test_max_length_pass(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(0), columns=Index(["value"]))
        check_pandas_dataframe(df, max_length=1)

    def test_max_length_error(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(2), columns=Index(["value"]))
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must satisfy the length requirements; got .*\n\n.*\.",
                flags=DOTALL,
            ),
        ):
            check_pandas_dataframe(df, max_length=1)

    def test_sorted_pass(self) -> None:
        df = DataFrame([[0.0], [1.0]], index=RangeIndex(2), columns=Index(["value"]))
        check_pandas_dataframe(df, sorted="value")

    def test_sorted_error(self) -> None:
        df = DataFrame([[1.0], [0.0]], index=RangeIndex(2), columns=Index(["value"]))
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(r"DataFrame must be sorted on .*\n\n.*\.", flags=DOTALL),
        ):
            check_pandas_dataframe(df, sorted="value")

    def test_standard_pass(self) -> None:
        check_pandas_dataframe(DataFrame(index=RangeIndex(0)), standard=True)

    @mark.parametrize(
        "df",
        [
            param(DataFrame(0.0, index=Index(["A"]), columns=Index(["value"]))),
            param(DataFrame(0.0, index=RangeIndex(1, 2), columns=Index(["value"]))),
            param(
                DataFrame(0.0, index=RangeIndex(1, step=2), columns=Index(["value"]))
            ),
            param(
                DataFrame(
                    0.0, index=RangeIndex(1, name="name"), columns=Index(["value"])
                )
            ),
        ],
    )
    def test_standard_errors_index(self, *, df: DataFrame) -> None:
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must have a standard index; got .*\n\n.*\.", flags=DOTALL
            ),
        ):
            check_pandas_dataframe(df, standard=True)

    @mark.parametrize(
        "df",
        [
            param(
                DataFrame(
                    0.0, index=RangeIndex(1), columns=Index(["value"], name="name")
                )
            ),
            param(
                DataFrame(0.0, index=RangeIndex(1), columns=Index(["value", "value"]))
            ),
        ],
    )
    def test_standard_errors(self, *, df: DataFrame) -> None:
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must have standard columns; got .*\n\n.*\.", flags=DOTALL
            ),
        ):
            check_pandas_dataframe(df, standard=True)

    def test_unique_pass(self) -> None:
        df = DataFrame([[0.0], [1.0]], index=RangeIndex(2), columns=Index(["value"]))
        check_pandas_dataframe(df, unique="value")

    def test_unique_error(self) -> None:
        df = DataFrame(0.0, index=RangeIndex(2), columns=Index(["value"]))
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(r"DataFrame must be unique on .*\n\n.*\.", flags=DOTALL),
        ):
            check_pandas_dataframe(df, unique="value")

    def test_width_pass(self) -> None:
        df = DataFrame()
        check_pandas_dataframe(df, width=0)

    def test_width_error(self) -> None:
        df = DataFrame()
        with raises(
            CheckPandasDataFrameError,
            match=re.compile(
                r"DataFrame must have width .*; got .*\n\n.*\.", flags=DOTALL
            ),
        ):
            check_pandas_dataframe(df, width=1)


class TestCheckRangeIndex:
    def test_main(self) -> None:
        check_range_index(RangeIndex(0))

    def test_start_pass(self) -> None:
        check_range_index(RangeIndex(0), start=0)

    def test_start_error(self) -> None:
        with raises(CheckRangeIndexError):
            check_range_index(RangeIndex(0), start=1)

    def test_stop_pass(self) -> None:
        check_range_index(RangeIndex(0), stop=0)

    def test_stop_error(self) -> None:
        with raises(CheckRangeIndexError):
            check_range_index(RangeIndex(0), stop=1)

    def test_step_pass(self) -> None:
        check_range_index(RangeIndex(0), step=1)

    def test_step_error(self) -> None:
        with raises(CheckRangeIndexError):
            check_range_index(RangeIndex(0), step=2)

    def test_length_pass(self) -> None:
        check_range_index(RangeIndex(1), length=1)

    def test_length_error(self) -> None:
        with raises(CheckRangeIndexError):
            check_range_index(RangeIndex(1), length=2)

    def test_min_length_pass(self) -> None:
        check_range_index(RangeIndex(2), min_length=1)

    def test_min_length_error(self) -> None:
        with raises(CheckRangeIndexError):
            check_range_index(RangeIndex(0), min_length=1)

    def test_max_length_pass(self) -> None:
        check_range_index(RangeIndex(0), max_length=1)

    def test_max_length_error(self) -> None:
        with raises(CheckRangeIndexError):
            check_range_index(RangeIndex(2), max_length=1)

    def test_name_pass(self) -> None:
        check_range_index(RangeIndex(0), name=None)

    def test_name_error(self) -> None:
        with raises(CheckRangeIndexError):
            check_range_index(RangeIndex(0), name="name")


class TestDTypes:
    @mark.parametrize("dtype", [param(Int64), param(boolean), param(string)])
    def test_main(self, *, dtype: Any) -> None:
        assert isinstance(Series([], dtype=dtype), Series)


class TestReindexToSet:
    @given(name=text_ascii() | none())
    def test_main(self, *, name: str | None) -> None:
        index = Index([1, 2, 3], name=name)
        target = [3, 2, 1]
        result = reindex_to_set(index, target)
        expected = Index([3, 2, 1], name=name)
        assert_index_equal(result, expected)

    def test_error(self) -> None:
        index = Index([1, 2, 3])
        target = [2, 3, 4]
        with raises(
            ReindexToSetError, match=r"Index .* and .* must be equal as sets\."
        ):
            _ = reindex_to_set(index, target)


class TestReindexToSubSet:
    @given(name=text_ascii() | none())
    def test_main(self, *, name: str | None) -> None:
        index = Index([1, 2, 3], name=name)
        target = [1]
        result = reindex_to_subset(index, target)
        expected = Index([1], name=name)
        assert_index_equal(result, expected)

    def test_error(self) -> None:
        index = Index([1])
        target = [1, 2, 3]
        with raises(ReindexToSubSetError, match=r"Index .* must be a superset of .*\."):
            _ = reindex_to_subset(index, target)


class TestReindexToSuperSet:
    @given(name=text_ascii() | none())
    def test_main(self, *, name: str | None) -> None:
        index = Index([1], name=name)
        target = [1, 2, 3]
        result = reindex_to_superset(index, target)
        expected = Index([1, 2, 3], name=name)
        assert_index_equal(result, expected)

    def test_error(self) -> None:
        index = Index([1, 2, 3])
        target = [1]
        with raises(ReindexToSuperSetError, match=r"Index .* must be a subset of .*\."):
            _ = reindex_to_superset(index, target)


class TestRedirectEmptyPandasConcat:
    def test_main(self) -> None:
        with raises(EmptyPandasConcatError), redirect_empty_pandas_concat():
            _ = concat([])


class TestSeriesMinMax:
    @mark.parametrize(
        ("x_v", "y_v", "dtype", "expected_min_v", "expected_max_v"),
        [
            param(0.0, 1.0, float, 0.0, 1.0),
            param(0.0, nan, float, 0.0, 0.0),
            param(nan, 1.0, float, 1.0, 1.0),
            param(nan, nan, float, nan, nan),
            param(0, 1, Int64, 0, 1),
            param(0, NA, Int64, 0, 0),
            param(NA, 1, Int64, 1, 1),
            param(NA, NA, Int64, NA, NA),
            param(
                TIMESTAMP_MIN_AS_DATE,
                TIMESTAMP_MAX_AS_DATE,
                datetime64ns,
                TIMESTAMP_MIN_AS_DATE,
                TIMESTAMP_MAX_AS_DATE,
            ),
            param(
                TIMESTAMP_MIN_AS_DATE,
                NaT,
                datetime64ns,
                TIMESTAMP_MIN_AS_DATE,
                TIMESTAMP_MIN_AS_DATE,
            ),
            param(
                NaT,
                TIMESTAMP_MAX_AS_DATE,
                datetime64ns,
                TIMESTAMP_MAX_AS_DATE,
                TIMESTAMP_MAX_AS_DATE,
            ),
            param(NaT, NaT, datetime64ns, NaT, NaT),
        ],
    )
    def test_main(
        self,
        *,
        x_v: Any,
        y_v: Any,
        dtype: Any,
        expected_min_v: Any,
        expected_max_v: Any,
    ) -> None:
        x = Series(data=[x_v], dtype=dtype)
        y = Series(data=[y_v], dtype=dtype)
        result_min = series_min(x, y)
        expected_min = Series(data=[expected_min_v], dtype=dtype)
        assert_series_equal(result_min, expected_min)
        result_max = series_max(x, y)
        expected_max = Series(data=[expected_max_v], dtype=dtype)
        assert_series_equal(result_max, expected_max)

    @mark.parametrize("func", [param(series_min), param(series_max)])
    def test_different_index(
        self, *, func: Callable[[SeriesA, SeriesA], SeriesA]
    ) -> None:
        x = Series(data=nan, index=Index([0], dtype=int))
        y = Series(data=nan, index=Index([1], dtype=int))
        with raises(AssertionError):
            _ = func(x, y)

    @mark.parametrize("func", [param(series_min), param(series_max)])
    def test_error(self, *, func: Callable[[SeriesA, SeriesA], SeriesA]) -> None:
        x = Series(data=nan, dtype=float)
        y = Series(data=NA, dtype=Int64)
        with raises(
            SeriesMinMaxError,
            match=re.compile(
                r"Series .* and .* must have the same dtype; got .* and .*\.",
                flags=DOTALL,
            ),
        ):
            _ = func(x, y)


class TestTimestampMinMaxAsDate:
    def test_min(self) -> None:
        date = TIMESTAMP_MIN_AS_DATE
        assert isinstance(to_datetime(cast(Timestamp, date)), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(cast(Timestamp, date - dt.timedelta(days=1)))

    def test_max(self) -> None:
        date = TIMESTAMP_MAX_AS_DATE
        assert isinstance(to_datetime(cast(Timestamp, date)), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(cast(Timestamp, date + dt.timedelta(days=1)))


class TestTimestampMinMaxAsDateTime:
    def test_min(self) -> None:
        date = TIMESTAMP_MIN_AS_DATETIME
        assert isinstance(to_datetime(date), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(date - dt.timedelta(microseconds=1))

    def test_max(self) -> None:
        date = TIMESTAMP_MAX_AS_DATETIME
        assert isinstance(to_datetime(date), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(date + dt.timedelta(microseconds=1))


class TestTimestampToDate:
    @mark.parametrize(
        ("timestamp", "expected"),
        [
            param(to_datetime("2000-01-01"), dt.date(2000, 1, 1)),
            param(to_datetime("2000-01-01 12:00:00"), dt.date(2000, 1, 1)),
        ],
    )
    def test_main(self, *, timestamp: Any, expected: dt.date) -> None:
        assert timestamp_to_date(timestamp) == expected

    def test_error(self) -> None:
        with raises(TimestampToDateTimeError):
            _ = timestamp_to_date(NaT)


class TestTimestampToDateTime:
    @mark.parametrize(
        ("timestamp", "expected"),
        [
            param(to_datetime("2000-01-01"), dt.datetime(2000, 1, 1, tzinfo=UTC)),
            param(
                to_datetime("2000-01-01 12:00:00"),
                dt.datetime(2000, 1, 1, 12, tzinfo=UTC),
            ),
            param(
                to_datetime("2000-01-01 12:00:00+00:00"),
                dt.datetime(2000, 1, 1, 12, tzinfo=UTC),
            ),
        ],
    )
    def test_main(self, *, timestamp: Any, expected: dt.datetime) -> None:
        assert timestamp_to_datetime(timestamp) == expected

    @given(timestamp=timestamps(allow_nanoseconds=True))
    def test_warn(self, *, timestamp: Timestamp) -> None:
        _ = assume(cast(Any, timestamp).nanosecond != 0)
        with raises(UserWarning, match="Discarding nonzero nanoseconds in conversion"):
            _ = timestamp_to_datetime(timestamp)

    def test_error(self) -> None:
        with raises(TimestampToDateTimeError):
            _ = timestamp_to_datetime(NaT)


class TestToNumpy:
    @mark.parametrize(
        ("series_v", "series_d", "array_v", "array_d"),
        [
            param(True, bool, True, bool),
            param(False, bool, False, bool),
            param(True, boolean, True, object),
            param(False, boolean, False, object),
            param(NA, boolean, None, object),
            param(TODAY_UTC, datetime64ns, TODAY_UTC, datetime64ns),
            param(0, int, 0, int),
            param(0, Int64, 0, object),
            param(NA, Int64, None, object),
            param(nan, float, nan, float),
            param("", string, "", object),
            param(NA, string, None, object),
        ],
    )
    def test_main(
        self, *, series_v: Any, series_d: Any, array_v: Any, array_d: Any
    ) -> None:
        series = Series([series_v], dtype=series_d)
        result = to_numpy(series)
        expected = array([array_v], dtype=array_d)
        assert_equal(result, expected)


class TestUnionIndexes:
    @given(name=text_ascii() | none())
    def test_first_named(self, *, name: str | None) -> None:
        left = Index([1, 2, 3], name=name)
        right = Index([2, 3, 4])
        result1 = union_indexes(left, right)
        result2 = union_indexes(right, left)
        expected = Index([1, 2, 3, 4], name=name)
        assert_index_equal(result1, expected)
        assert_index_equal(result2, expected)

    @given(lname=text_ascii(), rname=text_ascii())
    def test_both_named_taking_first(self, *, lname: str, rname: str) -> None:
        left = Index([1, 2, 3], name=lname)
        right = Index([2, 3, 4], name=rname)
        result = union_indexes(left, right, names="first")
        expected = Index([1, 2, 3, 4], name=lname)
        assert_index_equal(result, expected)

    @given(lname=text_ascii(), rname=text_ascii())
    def test_both_named_taking_last(self, *, lname: str, rname: str) -> None:
        left = Index([1, 2, 3], name=lname)
        right = Index([2, 3, 4], name=rname)
        result = union_indexes(left, right, names="last")
        expected = Index([1, 2, 3, 4], name=rname)
        assert_index_equal(result, expected)

    @given(lname=text_ascii(), rname=text_ascii())
    def test_both_named_error(self, *, lname: str, rname: str) -> None:
        _ = assume(lname != rname)
        left = Index([1, 2, 3], name=lname)
        right = Index([2, 3, 4], name=rname)
        with raises(
            UnionIndexesError,
            match=r"Indexes .* and .* must have the same name; got .* and .*\.",
        ):
            _ = union_indexes(left, right, names="raise")



from __future__ import annotations

from typing import TYPE_CHECKING

from click.testing import CliRunner
from hypothesis import given
from hypothesis.strategies import integers

from utilities.hypothesis import text_ascii
from utilities.pathlib import temp_cwd
from utilities.scripts.luigi_server import _get_args, main

if TYPE_CHECKING:
    from pathlib import Path


class TestLuigiServer:
    @given(
        pid_file=text_ascii(min_size=1),
        log_dir=text_ascii(min_size=1),
        state_path=text_ascii(min_size=1),
        port=integers(),
    )
    def test_get_args(
        self, *, pid_file: Path, log_dir: str, state_path: str, port: int
    ) -> None:
        _ = _get_args(
            pid_file=pid_file, log_dir=log_dir, state_path=state_path, port=port
        )

    def test_dry_run(self, *, tmp_path: Path) -> None:
        runner = CliRunner()
        args = ["--dry-run"]
        with temp_cwd(tmp_path):
            result = runner.invoke(main, args)
        assert result.exit_code == 0
